# Implementation

This chapter details the implementation of a comprehensive benchmarking
framework to evaluate several image processing libraries, including
ImageSharp, OpenCvSharp paired with SkiaSharp, Emgu CV coupled with
Structure.Sketching, and Magick.NET integrated with MagicScaler. The
objective was to create an end‐to‐end system that not only measures
execution times for common image operations but also provides insights
into memory usage.

This has been sought to answer key questions regarding the efficiency of
image conversion and pixel iteration operations—two fundamental tasks in
image processing. The following sections describe the review process,
architectural decisions, and technical implementations in the study. The
full implementation, including source code and benchmarking results, is
available at Gitlab repository[^1].

## System Architecture and Design Rationale

The design of the benchmarking framework was guided by the need for
consistency, repeatability, and scientific severity. The system was
architected to support multiple libraries through a common interface,
ensuring that each library’s performance could be measured under
identical conditions. At the core of the design was a two‐phase
benchmarking process: an initial warm-up phase to account for any
initialization overhead, followed by a main test phase where the actual
performance metrics were recorded.

In constructing the system, several important decisions were made.
First, a modular approach was employed, separating the benchmarking
routines into distinct components. This allowed the logic for image
conversion and pixel iteration to be encapsulated into separate classes,
each responsible for executing a series of timed iterations and logging
the results.

``` [Sharp]C
public class ImageConversionBenchmark{
        
        // Benchmarking logic for image conversion
    }
    public class PixelIterationBenchmark{

        // Benchmarking logic for pixel iteration
    }
```

The architecture also included a dedicated component for result
aggregation, which exported data into an Excel file using EPPlus,
thereby facilitating further analysis and visualization.

``` [Sharp]C
using OfficeOpenXml;

    public class ExcelExporter{
        
        // Logic for exporting benchmark results to an Excel sheet in a structured format
    }
```

An essential aspect of the design was the uniformity of testing. Despite
the differences in methods of implementation among the libraries, the
benchmarking framework was designed to abstract away these differences.
Each library was integrated by implementing the same sequence of
operations: reading an image from disk, processing the image (either
converting its format or iterating over its pixels to apply a grayscale
filter), and finally saving the processed image back to disk. This
uniform methodology ensured that the performance comparisons were both
fair and reproducible.

The architecture also accounted for system-level factors such as memory
management and garbage collection. For instance, in languages like C#,
where unmanaged resources must be explicitly disposed of, the design
included rigorous cleanup routines to ensure that each iteration began
with a clean slate. This attention to detail was crucial in obtaining
accurate measurements, as any residual state from previous iterations
could skew the results.

``` [Sharp]C
using BenchmarkDotNet.Attributes;
    using BenchmarkDotNet.Running;

    class Program
    {
        static void Main(string[] args)
        {
            BenchmarkRunner.Run<Benchmarks>();
        }
    }

    [MemoryDiagnoser]
    public class Benchmarks{
        
        [Benchmark]
        public void ImageConversionBenchmark(){
            // Image conversion logic
        }

        [Benchmark]
        public void PixelIterationBenchmark(){
            // Pixel iteration logic
        }
    }
```

## Benchmarking Implementation

The implementation of the benchmarking framework is divided into two
main tests: the image conversion benchmark and the pixel iteration
benchmark. Both tests follow a similar structure, starting with a
warm-up phase to mitigate initialization effects, followed by a series
of iterations where performance metrics are recorded.

### Image Conversion Benchmark Implementation

The image conversion benchmark is designed to measure the time it takes
to load an image from disk, convert its format, and save the result.
This process is critical in many image processing pipelines, where quick
and efficient conversion between formats can significantly impact
overall throughput.

The code snippet below illustrates the core routine for this benchmark.
The process begins with a series of warm-up iterations, during which the
system’s just-in-time (JIT) compilation and caching mechanisms are
activated. After the warm-up phase, the main iterations are executed,
with each iteration logging the time taken for the conversion.

``` [Sharp]C
public class ImageConversionBenchmark
{
    public static (double warmupTime, double averageTime, double totalTime) RunBenchmark(string inputPath, string outputPath, int iterations)
    {
        long totalElapsedMilliseconds = 0;
        long warmupTime = 0;
        int warmupIterations = 5;
        Stopwatch stopwatch = new Stopwatch();

        // Warm-up iterations to allow the system to reach steady state.
        for (int i = 0; i < warmupIterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image image = Image.Load(inputPath))
            {
                using (FileStream fs = new FileStream(outputPath, FileMode.Create))
                {
                    image.Save(fs, new PngEncoder());
                }
            }
            stopwatch.Stop();
            warmupTime += stopwatch.ElapsedMilliseconds;
        }

        // Main iterations where actual performance data is collected.
        for (int i = 0; i < iterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image image = Image.Load(inputPath))
            {
                using (FileStream fs = new FileStream(outputPath, FileMode.Create))
                {
                    image.Save(fs, new PngEncoder());
                }
            }
            stopwatch.Stop();
            totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
            Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
        }

        double averageTime = totalElapsedMilliseconds / (double)iterations;
        double totalTime = warmupTime + totalElapsedMilliseconds;
        Console.WriteLine($"Warm-up: {warmupTime} ms, Average: {averageTime} ms, Total: {totalTime} ms");

        return (warmupTime, averageTime, totalTime);
    }
}
```

In the implementation, the warm-up phase executes for five iterations,
during which the image is loaded, converted to PNG format, and the
elapsed time is accumulated. Subsequently, the main benchmark conducts
100 iterations of the identical operation, enabling the calculation of a
statistically significant average execution time. This methodological
approach ensures the isolation of steady-state performance from
initialization overhead, providing metrics that accurately reflect the
operational cost of image conversion.

This benchmark design evolved through iterative refinement. Initial
experimental observations revealed significantly higher latency during
the first iterations, necessitating the introduction of a dedicated
warm-up phase. The benchmark methodology has been progressively
optimized to ensure maximum isolation between iterations, thereby
minimizing the influence of transient system states on measurement
accuracy.

While this benchmarking approach is primarily analytical in nature, it’s
implicit in practical applications as well. For persistent server
applications or batch processing systems, the steady-state performance
metrics post-warm-up represent the most relevant operational
characteristics. Conversely, for interactive applications, command-line
utilities, or serverless computing environments, the initial performance
represented by the warm-up phase may be more indicative of
user-perceived responsiveness, as these contexts typically experience
the full initialization cost with each invocation.

### Pixel Iteration Benchmark Implementation

The pixel iteration benchmark measures the time taken to perform a basic
image processing operation—converting an image to grayscale by iterating
over each pixel. While modern image processing often employs vectorized
operations on entire matrices for efficiency, pixel-by-pixel iteration
remains relevant in several scenarios: when implementing custom filters
with complex logic, when working with specialized pixel formats, or when
memory constraints limit bulk operations. Additionally, this benchmark
provides insight into the underlying performance characteristics of
image libraries even if vectorized alternatives would be preferred in
production environments. By examining the performance of this
fundamental operation, a better understanding of the efficiency
trade-offs in various image processing contexts is achieved.

For ImageSharp, the implementation involves loading the image as an
array of pixels, processing each pixel to compute its grayscale value,
and then updating the image accordingly. The following snippet provides
a glimpse into this process:

``` [Sharp]C
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.Formats.Png;
using SixLabors.ImageSharp.PixelFormats;
public class PixelIterationBenchmark
{
    public static (double warmupTime, double averageTime, double totalTime) RunBenchmark(string imagePath, int iterations)
    {
        long totalElapsedMilliseconds = 0;
        long warmupTime = 0;
        int warmupIterations = 5;
        Stopwatch stopwatch = new Stopwatch();

        // Warm-up phase for pixel iteration
        for (int i = 0; i < warmupIterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image<Rgba32> image = Image.Load<Rgba32>(imagePath))
            {
                int width = image.Width;
                int height = image.Height;
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        Rgba32 pixel = image[x, y];
                        byte gray = (byte)((pixel.R + pixel.G + pixel.B) / 3);
                        image[x, y] = new Rgba32(gray, gray, gray, pixel.A);
                    }
                }
            }
            stopwatch.Stop();
            warmupTime += stopwatch.ElapsedMilliseconds;
        }

        // Main iterations to measure pixel iteration performance
        for (int i = 0; i < iterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image<Rgba32> image = Image.Load<Rgba32>(imagePath))
            {
                int width = image.Width;
                int height = image.Height;
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        Rgba32 pixel = image[x, y];
                        byte gray = (byte)((pixel.R + pixel.G + pixel.B) / 3);
                        image[x, y] = new Rgba32(gray, gray, gray, pixel.A);
                    }
                }
            }
            stopwatch.Stop();
            totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
            Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
        }

        double averageTime = totalElapsedMilliseconds / (double)iterations;
        double totalTime = warmupTime + totalElapsedMilliseconds;
        Console.WriteLine($"Warm-up: {warmupTime} ms, Average: {averageTime} ms, Total: {totalTime} ms");

        return (warmupTime, averageTime, totalTime);
    }
}
```

The code measures the performance of a grayscale conversion operation by
iterating over each pixel of an image. As in the image conversion, it
uses a timer (Stopwatch) and divides the process into two phases: a
warm-up phase and a measurement phase. During the warm-up phase, the
image is loaded and processed five times. This phase helps stabilize
performance by mitigating any startup overheads. Each iteration involves
loading the image, iterating over its width and height, reading each
pixel, computing the grayscale value by averaging the red, green, and
blue channels, and assigning the new grayscale value back while
preserving the alpha channel. The use of the `using` statement ensures
that the image is properly disposed after processing.

In the measurement phase, the same processing occurs over a
user-specified number of iterations (100 iterations). After running all
iterations, the code calculates the average time per iteration and the
total time including warm-up. This approach isolates the steady-state
performance from any one-time overhead, resulting in more accurate
measurements that reflect the true cost of pixel-by-pixel manipulations.

The design emphasizes clear resource management, detailed timing, and
separation of initialization costs from the main measurement, which are
crucial when every microsecond of processing time matters in image
manipulation scenarios.

The main focus of the implementation was to capture the interplay
between algorithmic efficiency and system-level resource management.
Every pixel operation is executed in a closed loop, and even minor
inefficiencies can accumulate over hundreds of iterations. The loop
structure is designed and a stopwatch is used to measure elapsed time to
matter of attention that should be paid to details during development.
Because even in high-level libraries such as ImageSharp, every
microsecond counts when processing large images.

## Libraries Implementation

As discussed in the Methodology chapter, a comprehensive evaluation was
undertaken to assess the strengths and limitations of various image
processing libraries. This analysis informed the decision to implement
integrations for frameworks: OpenCvSharp with SkiaSharp, and Emgu CV
with Structure.Sketching, and Magick.NET with MagicScaler. The following
excerpt presents representative code segments that illustrate the
implementation strategies developed for these libraries. These segments
not only capture the theoretical rationale behind each implementation
approach but also reflect the practical constraints and performance
considerations addressed throughout the thesis.

### OpenCvSharp and SkiaSharp Implementation

The following implementation shows how the OpenCvSharp and SkiaSharp
libraries are integrated to perform image conversion and pixel iteration
tasks. Image conversion was implemented using OpenCvSharp, while pixel
iteration was implemented using SkiaSharp.

``` [Sharp]C
using OpenCvSharp;
using SkiaSharp;

// Image Conversion logic using SkiaSharp
public class ImageConversionBenchmark
{
 public static (double warmupTime, double averageTimeExcludingWarmup, double totalTimeIncludingWarmup) RunBenchmark(string inputPath, string outputPath, int iterations)
```

The ImageConversionBenchmark class contains a static method RunBenchmark
that takes the input image path, output image path, and number of
iterations as input parameters. The method returns a tuple containing
the warm-up time, average time excluding warm-up, and total time
including warm-up, which will be used to form the results Excel file.

``` [Sharp]C
{
 long totalElapsedMilliseconds = 0;
 long warmupTime = 0;
 int warmupIterations = 5;
 Stopwatch stopwatch = new Stopwatch();
```

First, the totalElapsedMilliseconds and warmupTime variables are
initialized and as discussed in the methodology chapter, the
warmupIterations are set to 5. A stopwatch object is created to measure
the elapsed time for each iteration.

``` [Sharp]C
// Warm-up iterations
 for (int i = 0; i < warmupIterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(inputPath, ImreadModes.Color))
 {
 Cv2.ImWrite(outputPath, image);
 }

 stopwatch.Stop();
 warmupTime += stopwatch.ElapsedMilliseconds;
 }
```

The warm-up phase is executed five times to ensure that the libraries
are fully initialized before the main iterations begin. In each
iteration, the code reads an image using `Cv2.ImRead`, and writes the
image using `Cv2.ImWrite`. The elapsed time for each iteration is
recorded using the stopwatch object.

``` [Sharp]C
// Main iterations
 for (int i = 0; i < iterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(inputPath, ImreadModes.Color))
 {
 Cv2.ImWrite(outputPath, image);
 }

 stopwatch.Stop();
 totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
 Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
 }
```

After the warm-up phase, the main iterations are executed, and the
elapsed time for each iteration is recorded. The results are then
aggregated and returned as a tuple containing the warm-up time, average
time excluding warm-up, and total time including warm-up.

``` [Sharp]C
double averageTimeExcludingWarmup = totalElapsedMilliseconds / (double)iterations;
 double totalTimeIncludingWarmup = warmupTime + totalElapsedMilliseconds;

 Console.WriteLine($"Warm-up time for image conversion: {warmupTime} ms");
 Console.WriteLine($"Average time excluding warm-up for image conversion: {averageTimeExcludingWarmup} ms");
 Console.WriteLine($"Total time including warm-up for image conversion: {totalTimeIncludingWarmup} ms");

 return (warmupTime, averageTimeExcludingWarmup, totalTimeIncludingWarmup);
 }
}
```

Finally, the average time excluding warm-up, total time including
warm-up, and warm-up time are calculated. These values are then printed
to the console and returned as a tuple containing the warm-up time,
average time excluding warm-up, and total time including warm-up.

The pixel iteration benchmark, on the other hand, uses SkiaSharp to
perform pixel-wise operations on the image.

Same as the image conversion benchmark, the pixel iteration benchmark is
implemented as a static method RunBenchmark that takes the image path
and the number of iterations as input parameters. The method returns a
tuple containing the warm-up time, average time excluding warm-up, and
total time including warm-up. And in the same way variables are
initialized.

``` [Sharp]C
// Warm-up iterations
 for (int i = 0; i < warmupIterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(imagePath, ImreadModes.Color))
 {
 for (int y = 0; y < image.Rows; y++)
 {
 for (int x = 0; x < image.Cols; x++)
 {
 var pixel = image.At<Vec3b>(y, x);
 byte gray = (byte)((pixel.Item0 + pixel.Item1 + pixel.Item2) / 3);
 image.Set(y, x, new Vec3b(gray, gray, gray));
 }
 }
 }

 stopwatch.Stop();
 warmupTime += stopwatch.ElapsedMilliseconds;
 }
```

The warm-up phase is executed five times to ensure that the libraries
are fully initialized before the main iterations begin. In each
iteration, the code reads an image using `Cv2.ImRead`, iterates over
each pixel, calculates the grayscale value, and then sets the pixel
value using `image.At<Vec3b>` and `image.Set`. The elapsed time for each
iteration is recorded using the stopwatch object.

``` [Sharp]C
// Main iterations
 for (int i = 0; i < iterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(imagePath, ImreadModes.Color))
 {
 for (int y = 0; y < image.Rows; y++)
 {
 for (int x = 0; x < image.Cols; x++)
 {
 var pixel = image.At<Vec3b>(y, x);
 byte gray = (byte)((pixel.Item0 + pixel.Item1 + pixel.Item2) / 3);
 image.Set(y, x, new Vec3b(gray, gray, gray));
 }
 }
 }

 stopwatch.Stop();
 totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
 Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
 }
```

After the warm-up phase, the main iterations are executed, using the
same logic as the warm-up phase. The elapsed time for each iteration is
recorded, and the results are then aggregated and returned as a tuple
containing the warm-up time, average time excluding warm-up, and total
time including warm-up.

``` [Sharp]C
double averageTimeExcludingWarmup = totalElapsedMilliseconds / (double)iterations;
 double totalTimeIncludingWarmup = warmupTime + totalElapsedMilliseconds;

 Console.WriteLine($"Warm-up time for pixel iteration: {warmupTime} ms");
 Console.WriteLine($"Average time excluding warm-up for pixel iteration: {averageTimeExcludingWarmup} ms");
 Console.WriteLine($"Total time including warm-up for pixel iteration: {totalTimeIncludingWarmup} ms");

 return (warmupTime, averageTimeExcludingWarmup, totalTimeIncludingWarmup);
 }
}
```

   

Finally, the average time excluding warm-up, total time including
warm-up, and warm-up time are calculated. These values are then printed
to the console and returned as a tuple containing the warm-up time,
average time excluding warm-up, and total time including warm-up. The
returned values are then used to generate the results in an Excel file.

### Magick.NET Implementation

In the implementation of both image conversion and pixel iteration
benchmarks, Magick.NET library was used. This decision was based on
Magick.NET’s comprehensive functionality, which includes support for
high-quality image conversion and efficient pixel-wise operations.  
Similar to the previous section on OpenCvSharp and SkiaSharp, the
ImageConversionBenchmark class for Magick.NET features a static
RunBenchmark method. In this method, the necessary variables are
initialized to measure and record the performance of image conversion
operations. This consistent approach across libraries facilitates a
clear comparison of their performance under similar conditions.  
In logic for the warm-up phase and main iterations, change was only the
library-specific functions used for image conversion and pixel
iteration. Implementing image conversion using Magick.NET involved
reading an image using `new MagickImage(inputPath)` to read an image and
`image.Write(outputPath, MagickFormat.Png)` to write an image, and the
image conversion benchmark was implemented.

``` [Sharp]C
// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(inputPath))
    {
        image.Write(outputPath, MagickFormat.Png);
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(inputPath))
    {
        image.Write(outputPath, MagickFormat.Png);
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
}
```

The pixel iteration benchmark was implemented by first retrieving the
pixel data using the `image.GetPixels()` method. Then, for each pixel,
the color channels were set to the same gray value using the
`pixels.SetPixel(x, y, new ushort[] { gray, gray, gray })` function.
This process was repeated for each pixel in the image for both the
warm-up phase and the main iterations.

``` [Sharp]C
// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(imagePath))
    {
        var pixels = image.GetPixels();
        for (int y = 0; y < image.Height; y++)
        {
            for (int x = 0; x < image.Width; x++)
            {
                var pixel = pixels.GetPixel(x, y); // Get pixel data
                ushort gray = (ushort)((pixel[0] + pixel[1] + pixel[2]) / 3); // Convert to grayscale
                pixels.SetPixel(x, y, new ushort[] { gray, gray, gray }); // Set pixel data with ushort[]
            }
        }
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(imagePath))
    {
        var pixels = image.GetPixels();
        for (int y = 0; y < image.Height; y++)
        {
            for (int x = 0; x < image.Width; x++)
            {
                var pixel = pixels.GetPixel(x, y); // Get pixel data
                ushort gray = (ushort)((pixel[0] + pixel[1] + pixel[2]) / 3); // Convert to grayscale
                pixels.SetPixel(x, y, new ushort[] { gray, gray, gray }); // Set pixel data with ushort[]
            }
        }
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
}
```

The results of the image conversion and pixel iteration benchmarks were
then like the previous libraries, aggregated and returned as a tuple
containing the warm-up time, average time excluding warm-up, and total
time including warm-up. These values were then used to generate the
results in an Excel file.

### Emgu CV and Structure.Sketching Implementation

The implementation of Emgu CV and Structure.Sketching libraries in the
benchmarking framework are shown in the following code snippet. The code
demonstrates how the Emgu CV library is used for image conversion, while
Structure.Sketching is used for pixel iteration.

For image conversion, the code reads an image using `CvInvoke.Imread`
and writes the image using `CvInvoke.Imwrite`. The warm-up phase and
main iterations are executed in a similar manner to the previous
libraries, with the elapsed time for each iteration recorded using a
stopwatch object.

``` [Sharp]C
using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using Structure.Sketching;
using Structure.Sketching.Formats;
using Structure.Sketching.Colors;

// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (Mat image = CvInvoke.Imread(inputPath, ImreadModes.Color))
    {
        CvInvoke.Imwrite(outputPath, image);
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (Mat image = CvInvoke.Imread(inputPath, ImreadModes.Color))
    {
        CvInvoke.Imwrite(outputPath, image);
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
}
```

For pixel iteration, it uses the Structure.Sketching and the code reads
an image using `new Structure.Sketching.Image(imagePath)` and iterates
over each pixel, calculating the grayscale value and setting the pixel
value using `image.Pixels[(y * width) + x]`. The warm-up phase and main
iterations are executed in a similar manner to the previous libraries,
with the elapsed time for each iteration recorded using a stopwatch
object.

``` [Sharp]C
// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    var image = new Structure.Sketching.Image(imagePath);
    int width = image.Width;
    int height = image.Height;

    for (int y = 0; y < height; y++)
    {
        for (int x = 0; x < width; x++)
        {
            var pixel = image.Pixels[(y * width) + x];
            byte gray = (byte)((pixel.Red + pixel.Green + pixel.Blue) / 3);
            image.Pixels[(y * width) + x] = new Color(gray, gray, gray, pixel.Alpha);
        }
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    var image = new Structure.Sketching.Image(imagePath);
    int width = image.Width;
    int height = image.Height;

    for (int y = 0; y < height; y++)
    {
        for (int x = 0; x < width; x++)
        {
            var pixel = image.Pixels[(y * width) + x];
            byte gray = (byte)((pixel.Red + pixel.Green + pixel.Blue) / 3);
            image.Pixels[(y * width) + x] = new Color(gray, gray, gray, pixel.Alpha);
        }
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
}
```

Grayscale conversion is performed on each pixel by computing the average
of the red, green, and blue components using the formula
`(byte)((pixel.Red + pixel.Green + pixel.Blue) / 3)`. The grayscale
value is then assigned to each color channel to create a grayscale
image. The benchmarking process collects the results from both image
conversion and pixel iteration. These results are aggregated into a
tuple containing the warm-up time, the average time (excluding the
warm-up phase), and the total time (including the warm-up phase).
Finally, this data is used to generate an Excel file that summarizes the
performance metrics.

## Memory Profiling and Performance Analysis

In any high-performance image processing application, it is not enough
to measure raw execution time; memory consumption is equally critical.
This section describes the integration of memory profiling into the
benchmarking framework to provide a comprehensive view of the
performance characteristics of each library and complement the
time-based measurements. Using BenchmarkDotNet—a powerful tool for .NET
performance analysis—detailed metrics on memory allocation and garbage
collection behavior were captured. This implementation allowed the
trade-offs between processing speed and resource utilization to be
better understood.

The memory profiling is designed to evaluate not only the mean execution
times but also the memory allocated during both image conversion and
pixel iteration tasks. Using BenchmarkDotNet’s `[MemoryDiagnoser]`,
`[Orderer]`, and `[RankColumn]` attributes, data on memory consumption,
garbage collection events, and total allocated memory were collected for
each benchmarked operation. The BenchmarkDotNet analyzer for each method
by default is configured to automatically determine how many warmup and
measurement iterations to run based on the workload, environment, and
statistical requirements for accurate measurements. So there is no need
to implement a fixed iteration count for each method manually.

The following framework demonstrates the implementation of memory
profiling and an example of how the memory diagnostics were implemented
for the image conversion and pixel iteration using ImageSharp:

``` [Sharp]C
using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Order;
using BenchmarkDotNet.Running;
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.Formats.Png;
using SixLabors.ImageSharp.PixelFormats;

[MemoryDiagnoser]
[Orderer(SummaryOrderPolicy.FastestToSlowest)]
[RankColumn]
public class Benchmarks
{
    private const string InputImagePath = "./../../../../../xl1.jpg";
    private const string OutputImagePath = "./../../../../o.png";

    [Benchmark]
    public void ImageConversionBenchmark()
    {
        using (Image image = Image.Load(InputImagePath))
        {
            using (FileStream fs = new FileStream(OutputImagePath, FileMode.Create))
            {
                image.Save(fs, new PngEncoder());
                Console.WriteLine("ImageConversionBenchmark completed");
            }
        }
    }
```

Same logic is used for image conversion, but there were no need for
iterations and warm-up phase to be implemented manually. For configuring
the `MemoryDiagnoser` results,
`Orderer(SummaryOrderPolicy.FastestToSlowest)` and `RankColumn`
attributes were used to order the results based on the fastest to
slowest execution times and to rank the results in the summary table,
respectively to provide a better and clearer view of the results.

``` [Sharp]C
[Benchmark]
public void PixelIterationBenchmark()
{
    using (Image<Rgba32> image = Image.Load<Rgba32>(InputImagePath))
    {
        int width = image.Width;
        int height = image.Height;

        for (int y = 0; y < height; y++)
        {
            for (int x = 0; x < width; x++)
            {
                Rgba32 pixel = image[x, y];
                byte gray = (byte)((pixel.R + pixel.G + pixel.B) / 3);
                image[x, y] = new Rgba32(gray, gray, gray, pixel.A);
            }
        }
        Console.WriteLine("PixelIterationBenchmark completed");
    }
}
```

The pixel iteration benchmark was implemented in a similar manner with
the same memory diagnostics attributes. The code snippet above
demonstrates the pixel iteration benchmark for ImageSharp, where each
pixel in the image is converted to grayscale. The memory diagnostics
provided by BenchmarkDotNet enabled tracking of the memory consumption
and garbage collection events during the pixel iteration operation,
providing valuable insights into the resource utilization of each
library.

This code exemplifies the approach to memory diagnostics. By annotating
the benchmark class with `[MemoryDiagnoser]`, BenchmarkDotNet
automatically collects data on memory usage—including the number of
garbage collection (GC) events and the total allocated memory during
each benchmarked operation. Similar implementations were done for other
libraries as well.

This level of granularity provided insights that went beyond raw timing
metrics, revealing, for example, that while Emgu CV might be faster in
certain operations, its higher memory consumption could be a concern for
applications running on memory-constrained systems.

## Result Export and Data Aggregation

Once the performance and memory metrics were collected, the next
challenge was to present the results in a coherent and accessible
manner. For this purpose, Excel was chosen as the output format due to
its widespread adoption and ease of use for further analysis.
`OfficeOpenXml` namespace, which is part of the EPPlus library, allows
for the creation and manipulation of Excel files in .NET applications.
The ExcelExporter class was implemented to aggregate the benchmark
results and export them to an Excel file.

The code snippet below illustrates how the benchmark results are
aggregated and exported to an Excel file:

``` [Sharp]C
using OfficeOpenXml;

public class ExcelExporter
{
    public static void ExportResults(string excelOutputPath, 
        (double warmupTime, double averageTime, double totalTime) imageConversionResults, 
        (double warmupTime, double averageTime, double totalTime) pixelIterationResults)
    {
        using (var package = new ExcelPackage())
        {
            var worksheet = package.Workbook.Worksheets.Add("Benchmark Results");
            worksheet.Cells[1, 1].Value = "Benchmark";
            worksheet.Cells[1, 2].Value = "Warm-Up Time (ms)";
            worksheet.Cells[1, 3].Value = "Average Time (ms)";
            worksheet.Cells[1, 4].Value = "Total Time (ms)";
            
            worksheet.Cells[2, 1].Value = "Image Conversion";
            worksheet.Cells[2, 2].Value = imageConversionResults.warmupTime;
            worksheet.Cells[2, 3].Value = imageConversionResults.averageTime;
            worksheet.Cells[2, 4].Value = imageConversionResults.totalTime;
            
            worksheet.Cells[3, 1].Value = "Pixel Iteration";
            worksheet.Cells[3, 2].Value = pixelIterationResults.warmupTime;
            worksheet.Cells[3, 3].Value = pixelIterationResults.averageTime;
            worksheet.Cells[3, 4].Value = pixelIterationResults.totalTime;
            
            package.SaveAs(new FileInfo(excelOutputPath));
        }
    }
}
```

The ExcelExporter class creates a structured Excel file with separate
sheets for each benchmark operation. The results are organized into
columns for the warm-up time, average time, and total time for each
operation. The resulting Excel file provides a clear and concise summary
of the benchmark results, making it easy to compare the performance and
memory characteristics of each library.

By automating the process of result aggregation, the framework not only
saves time but also minimizes the risk of manual errors. Each cell in
the generated Excel file is carefully populated with benchmark data, and
the resulting spreadsheet can be easily imported into analytical tools
for further exploration. This process of exporting results serves as a
bridge between the raw performance data and the actionable insights that
drive decision-making in software optimization.

[^1]: <https://mygit.th-deg.de/sf07627/fazeli_shahroudi-sepehr-master-sthesis>
