# Benchmarking Implementation

The implementation of the benchmarking framework is divided into two
main tests: the image conversion benchmark and the pixel iteration
benchmark. Both tests follow a similar structure, starting with a
warm-up phase to mitigate initialization effects, followed by a series
of iterations where performance metrics are recorded.

## Image Conversion Benchmark Implementation

The image conversion benchmark is designed to measure the time it takes
to load an image from disk, convert its format, and save the result.
This process is critical in many image processing pipelines, where quick
and efficient conversion between formats can significantly impact
overall throughput.

The code snippet below illustrates the core routine for this benchmark.
The process begins with a series of warm-up iterations, during which the
system’s just-in-time (JIT) compilation and caching mechanisms are
activated. After the warm-up phase, the main iterations are executed,
with each iteration logging the time taken for the conversion.

``` [Sharp]C
public class ImageConversionBenchmark
{
    public static (double warmupTime, double averageTime, double totalTime) RunBenchmark(string inputPath, string outputPath, int iterations)
    {
        long totalElapsedMilliseconds = 0;
        long warmupTime = 0;
        int warmupIterations = 5;
        Stopwatch stopwatch = new Stopwatch();

        // Warm-up iterations to allow the system to reach steady state.
        for (int i = 0; i < warmupIterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image image = Image.Load(inputPath))
            {
                using (FileStream fs = new FileStream(outputPath, FileMode.Create))
                {
                    image.Save(fs, new PngEncoder());
                }
            }
            stopwatch.Stop();
            warmupTime += stopwatch.ElapsedMilliseconds;
        }

        // Main iterations where actual performance data is collected.
        for (int i = 0; i < iterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image image = Image.Load(inputPath))
            {
                using (FileStream fs = new FileStream(outputPath, FileMode.Create))
                {
                    image.Save(fs, new PngEncoder());
                }
            }
            stopwatch.Stop();
            totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
            Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
        }

        double averageTime = totalElapsedMilliseconds / (double)iterations;
        double totalTime = warmupTime + totalElapsedMilliseconds;
        Console.WriteLine($"Warm-up: {warmupTime} ms, Average: {averageTime} ms, Total: {totalTime} ms");

        return (warmupTime, averageTime, totalTime);
    }
}
```

In the implementation, the warm-up phase executes for five iterations,
during which the image is loaded, converted to PNG format, and the
elapsed time is accumulated. Subsequently, the main benchmark conducts
100 iterations of the identical operation, enabling the calculation of a
statistically significant average execution time. This methodological
approach ensures the isolation of steady-state performance from
initialization overhead, providing metrics that accurately reflect the
operational cost of image conversion.

This benchmark design evolved through iterative refinement. Initial
experimental observations revealed significantly higher latency during
the first iterations, necessitating the introduction of a dedicated
warm-up phase. The benchmark methodology has been progressively
optimized to ensure maximum isolation between iterations, thereby
minimizing the influence of transient system states on measurement
accuracy.

While this benchmarking approach is primarily analytical in nature, it’s
implicit in practical applications as well. For persistent server
applications or batch processing systems, the steady-state performance
metrics post-warm-up represent the most relevant operational
characteristics. Conversely, for interactive applications, command-line
utilities, or serverless computing environments, the initial performance
represented by the warm-up phase may be more indicative of
user-perceived responsiveness, as these contexts typically experience
the full initialization cost with each invocation.
