# Libraries Implementation

As discussed in the Methodology chapter, a comprehensive evaluation was
undertaken to assess the strengths and limitations of various image
processing libraries. This analysis informed the decision to implement
integrations for frameworks: OpenCvSharp with SkiaSharp, and Emgu CV
with Structure.Sketching, and Magick.NET with MagicScaler. The following
excerpt presents representative code segments that illustrate the
implementation strategies developed for these libraries. These segments
not only capture the theoretical rationale behind each implementation
approach but also reflect the practical constraints and performance
considerations addressed throughout the thesis.

## OpenCvSharp and SkiaSharp Implementation

The following implementation shows how the OpenCvSharp and SkiaSharp
libraries are integrated to perform image conversion and pixel iteration
tasks. Image conversion was implemented using OpenCvSharp, while pixel
iteration was implemented using SkiaSharp.

``` [Sharp]C
using OpenCvSharp;
using SkiaSharp;

// Image Conversion logic using SkiaSharp
public class ImageConversionBenchmark
{
 public static (double warmupTime, double averageTimeExcludingWarmup, double totalTimeIncludingWarmup) RunBenchmark(string inputPath, string outputPath, int iterations)
```

The ImageConversionBenchmark class contains a static method RunBenchmark
that takes the input image path, output image path, and number of
iterations as input parameters. The method returns a tuple containing
the warm-up time, average time excluding warm-up, and total time
including warm-up, which will be used to form the results Excel file.

``` [Sharp]C
{
 long totalElapsedMilliseconds = 0;
 long warmupTime = 0;
 int warmupIterations = 5;
 Stopwatch stopwatch = new Stopwatch();
```

First, the totalElapsedMilliseconds and warmupTime variables are
initialized and as discussed in the methodology chapter, the
warmupIterations are set to 5. A stopwatch object is created to measure
the elapsed time for each iteration.

``` [Sharp]C
// Warm-up iterations
 for (int i = 0; i < warmupIterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(inputPath, ImreadModes.Color))
 {
 Cv2.ImWrite(outputPath, image);
 }

 stopwatch.Stop();
 warmupTime += stopwatch.ElapsedMilliseconds;
 }
```

The warm-up phase is executed five times to ensure that the libraries
are fully initialized before the main iterations begin. In each
iteration, the code reads an image using `Cv2.ImRead`, and writes the
image using `Cv2.ImWrite`. The elapsed time for each iteration is
recorded using the stopwatch object.

``` [Sharp]C
// Main iterations
 for (int i = 0; i < iterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(inputPath, ImreadModes.Color))
 {
 Cv2.ImWrite(outputPath, image);
 }

 stopwatch.Stop();
 totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
 Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
 }
```

After the warm-up phase, the main iterations are executed, and the
elapsed time for each iteration is recorded. The results are then
aggregated and returned as a tuple containing the warm-up time, average
time excluding warm-up, and total time including warm-up.

``` [Sharp]C
double averageTimeExcludingWarmup = totalElapsedMilliseconds / (double)iterations;
 double totalTimeIncludingWarmup = warmupTime + totalElapsedMilliseconds;

 Console.WriteLine($"Warm-up time for image conversion: {warmupTime} ms");
 Console.WriteLine($"Average time excluding warm-up for image conversion: {averageTimeExcludingWarmup} ms");
 Console.WriteLine($"Total time including warm-up for image conversion: {totalTimeIncludingWarmup} ms");

 return (warmupTime, averageTimeExcludingWarmup, totalTimeIncludingWarmup);
 }
}
```

Finally, the average time excluding warm-up, total time including
warm-up, and warm-up time are calculated. These values are then printed
to the console and returned as a tuple containing the warm-up time,
average time excluding warm-up, and total time including warm-up.

The pixel iteration benchmark, on the other hand, uses SkiaSharp to
perform pixel-wise operations on the image.

Same as the image conversion benchmark, the pixel iteration benchmark is
implemented as a static method RunBenchmark that takes the image path
and the number of iterations as input parameters. The method returns a
tuple containing the warm-up time, average time excluding warm-up, and
total time including warm-up. And in the same way variables are
initialized.

``` [Sharp]C
// Warm-up iterations
 for (int i = 0; i < warmupIterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(imagePath, ImreadModes.Color))
 {
 for (int y = 0; y < image.Rows; y++)
 {
 for (int x = 0; x < image.Cols; x++)
 {
 var pixel = image.At<Vec3b>(y, x);
 byte gray = (byte)((pixel.Item0 + pixel.Item1 + pixel.Item2) / 3);
 image.Set(y, x, new Vec3b(gray, gray, gray));
 }
 }
 }

 stopwatch.Stop();
 warmupTime += stopwatch.ElapsedMilliseconds;
 }
```

The warm-up phase is executed five times to ensure that the libraries
are fully initialized before the main iterations begin. In each
iteration, the code reads an image using `Cv2.ImRead`, iterates over
each pixel, calculates the grayscale value, and then sets the pixel
value using `image.At<Vec3b>` and `image.Set`. The elapsed time for each
iteration is recorded using the stopwatch object.

``` [Sharp]C
// Main iterations
 for (int i = 0; i < iterations; i++)
 {
 stopwatch.Reset();
 stopwatch.Start();

 using (var image = Cv2.ImRead(imagePath, ImreadModes.Color))
 {
 for (int y = 0; y < image.Rows; y++)
 {
 for (int x = 0; x < image.Cols; x++)
 {
 var pixel = image.At<Vec3b>(y, x);
 byte gray = (byte)((pixel.Item0 + pixel.Item1 + pixel.Item2) / 3);
 image.Set(y, x, new Vec3b(gray, gray, gray));
 }
 }
 }

 stopwatch.Stop();
 totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
 Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
 }
```

After the warm-up phase, the main iterations are executed, using the
same logic as the warm-up phase. The elapsed time for each iteration is
recorded, and the results are then aggregated and returned as a tuple
containing the warm-up time, average time excluding warm-up, and total
time including warm-up.

``` [Sharp]C
double averageTimeExcludingWarmup = totalElapsedMilliseconds / (double)iterations;
 double totalTimeIncludingWarmup = warmupTime + totalElapsedMilliseconds;

 Console.WriteLine($"Warm-up time for pixel iteration: {warmupTime} ms");
 Console.WriteLine($"Average time excluding warm-up for pixel iteration: {averageTimeExcludingWarmup} ms");
 Console.WriteLine($"Total time including warm-up for pixel iteration: {totalTimeIncludingWarmup} ms");

 return (warmupTime, averageTimeExcludingWarmup, totalTimeIncludingWarmup);
 }
}
```

   

Finally, the average time excluding warm-up, total time including
warm-up, and warm-up time are calculated. These values are then printed
to the console and returned as a tuple containing the warm-up time,
average time excluding warm-up, and total time including warm-up. The
returned values are then used to generate the results in an Excel file.

## Magick.NET Implementation

In the implementation of both image conversion and pixel iteration
benchmarks, Magick.NET library was used. This decision was based on
Magick.NET’s comprehensive functionality, which includes support for
high-quality image conversion and efficient pixel-wise operations.  
Similar to the previous section on OpenCvSharp and SkiaSharp, the
ImageConversionBenchmark class for Magick.NET features a static
RunBenchmark method. In this method, the necessary variables are
initialized to measure and record the performance of image conversion
operations. This consistent approach across libraries facilitates a
clear comparison of their performance under similar conditions.  
In logic for the warm-up phase and main iterations, change was only the
library-specific functions used for image conversion and pixel
iteration. Implementing image conversion using Magick.NET involved
reading an image using `new MagickImage(inputPath)` to read an image and
`image.Write(outputPath, MagickFormat.Png)` to write an image, and the
image conversion benchmark was implemented.

``` [Sharp]C
// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(inputPath))
    {
        image.Write(outputPath, MagickFormat.Png);
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(inputPath))
    {
        image.Write(outputPath, MagickFormat.Png);
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
}
```

The pixel iteration benchmark was implemented by first retrieving the
pixel data using the `image.GetPixels()` method. Then, for each pixel,
the color channels were set to the same gray value using the
`pixels.SetPixel(x, y, new ushort[] { gray, gray, gray })` function.
This process was repeated for each pixel in the image for both the
warm-up phase and the main iterations.

``` [Sharp]C
// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(imagePath))
    {
        var pixels = image.GetPixels();
        for (int y = 0; y < image.Height; y++)
        {
            for (int x = 0; x < image.Width; x++)
            {
                var pixel = pixels.GetPixel(x, y); // Get pixel data
                ushort gray = (ushort)((pixel[0] + pixel[1] + pixel[2]) / 3); // Convert to grayscale
                pixels.SetPixel(x, y, new ushort[] { gray, gray, gray }); // Set pixel data with ushort[]
            }
        }
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (var image = new MagickImage(imagePath))
    {
        var pixels = image.GetPixels();
        for (int y = 0; y < image.Height; y++)
        {
            for (int x = 0; x < image.Width; x++)
            {
                var pixel = pixels.GetPixel(x, y); // Get pixel data
                ushort gray = (ushort)((pixel[0] + pixel[1] + pixel[2]) / 3); // Convert to grayscale
                pixels.SetPixel(x, y, new ushort[] { gray, gray, gray }); // Set pixel data with ushort[]
            }
        }
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
}
```

The results of the image conversion and pixel iteration benchmarks were
then like the previous libraries, aggregated and returned as a tuple
containing the warm-up time, average time excluding warm-up, and total
time including warm-up. These values were then used to generate the
results in an Excel file.

## Emgu CV and Structure.Sketching Implementation

The implementation of Emgu CV and Structure.Sketching libraries in the
benchmarking framework are shown in the following code snippet. The code
demonstrates how the Emgu CV library is used for image conversion, while
Structure.Sketching is used for pixel iteration.

For image conversion, the code reads an image using `CvInvoke.Imread`
and writes the image using `CvInvoke.Imwrite`. The warm-up phase and
main iterations are executed in a similar manner to the previous
libraries, with the elapsed time for each iteration recorded using a
stopwatch object.

``` [Sharp]C
using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using Structure.Sketching;
using Structure.Sketching.Formats;
using Structure.Sketching.Colors;

// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (Mat image = CvInvoke.Imread(inputPath, ImreadModes.Color))
    {
        CvInvoke.Imwrite(outputPath, image);
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    using (Mat image = CvInvoke.Imread(inputPath, ImreadModes.Color))
    {
        CvInvoke.Imwrite(outputPath, image);
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Image conversion took {stopwatch.ElapsedMilliseconds} ms");
}
```

For pixel iteration, it uses the Structure.Sketching and the code reads
an image using `new Structure.Sketching.Image(imagePath)` and iterates
over each pixel, calculating the grayscale value and setting the pixel
value using `image.Pixels[(y * width) + x]`. The warm-up phase and main
iterations are executed in a similar manner to the previous libraries,
with the elapsed time for each iteration recorded using a stopwatch
object.

``` [Sharp]C
// Warm-up iterations
for (int i = 0; i < warmupIterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    var image = new Structure.Sketching.Image(imagePath);
    int width = image.Width;
    int height = image.Height;

    for (int y = 0; y < height; y++)
    {
        for (int x = 0; x < width; x++)
        {
            var pixel = image.Pixels[(y * width) + x];
            byte gray = (byte)((pixel.Red + pixel.Green + pixel.Blue) / 3);
            image.Pixels[(y * width) + x] = new Color(gray, gray, gray, pixel.Alpha);
        }
    }

    stopwatch.Stop();
    warmupTime += stopwatch.ElapsedMilliseconds;
}

// Main iterations
for (int i = 0; i < iterations; i++)
{
    stopwatch.Reset();
    stopwatch.Start();

    var image = new Structure.Sketching.Image(imagePath);
    int width = image.Width;
    int height = image.Height;

    for (int y = 0; y < height; y++)
    {
        for (int x = 0; x < width; x++)
        {
            var pixel = image.Pixels[(y * width) + x];
            byte gray = (byte)((pixel.Red + pixel.Green + pixel.Blue) / 3);
            image.Pixels[(y * width) + x] = new Color(gray, gray, gray, pixel.Alpha);
        }
    }

    stopwatch.Stop();
    totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
    Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
}
```

Grayscale conversion is performed on each pixel by computing the average
of the red, green, and blue components using the formula
`(byte)((pixel.Red + pixel.Green + pixel.Blue) / 3)`. The grayscale
value is then assigned to each color channel to create a grayscale
image. The benchmarking process collects the results from both image
conversion and pixel iteration. These results are aggregated into a
tuple containing the warm-up time, the average time (excluding the
warm-up phase), and the total time (including the warm-up phase).
Finally, this data is used to generate an Excel file that summarizes the
performance metrics.
