# Memory Profiling and Performance Analysis

In any high-performance image processing application, it is not enough
to measure raw execution time; memory consumption is equally critical.
This section describes the integration of memory profiling into the
benchmarking framework to provide a comprehensive view of the
performance characteristics of each library and complement the
time-based measurements. Using BenchmarkDotNet—a powerful tool for .NET
performance analysis—detailed metrics on memory allocation and garbage
collection behavior were captured. This implementation allowed the
trade-offs between processing speed and resource utilization to be
better understood.

The memory profiling is designed to evaluate not only the mean execution
times but also the memory allocated during both image conversion and
pixel iteration tasks. Using BenchmarkDotNet’s `[MemoryDiagnoser]`,
`[Orderer]`, and `[RankColumn]` attributes, data on memory consumption,
garbage collection events, and total allocated memory were collected for
each benchmarked operation. The BenchmarkDotNet analyzer for each method
by default is configured to automatically determine how many warmup and
measurement iterations to run based on the workload, environment, and
statistical requirements for accurate measurements. So there is no need
to implement a fixed iteration count for each method manually.

The following framework demonstrates the implementation of memory
profiling and an example of how the memory diagnostics were implemented
for the image conversion and pixel iteration using ImageSharp:

``` [Sharp]C
using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Order;
using BenchmarkDotNet.Running;
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.Formats.Png;
using SixLabors.ImageSharp.PixelFormats;

[MemoryDiagnoser]
[Orderer(SummaryOrderPolicy.FastestToSlowest)]
[RankColumn]
public class Benchmarks
{
    private const string InputImagePath = "./../../../../../xl1.jpg";
    private const string OutputImagePath = "./../../../../o.png";

    [Benchmark]
    public void ImageConversionBenchmark()
    {
        using (Image image = Image.Load(InputImagePath))
        {
            using (FileStream fs = new FileStream(OutputImagePath, FileMode.Create))
            {
                image.Save(fs, new PngEncoder());
                Console.WriteLine("ImageConversionBenchmark completed");
            }
        }
    }
```

Same logic is used for image conversion, but there were no need for
iterations and warm-up phase to be implemented manually. For configuring
the `MemoryDiagnoser` results,
`Orderer(SummaryOrderPolicy.FastestToSlowest)` and `RankColumn`
attributes were used to order the results based on the fastest to
slowest execution times and to rank the results in the summary table,
respectively to provide a better and clearer view of the results.

``` [Sharp]C
[Benchmark]
public void PixelIterationBenchmark()
{
    using (Image<Rgba32> image = Image.Load<Rgba32>(InputImagePath))
    {
        int width = image.Width;
        int height = image.Height;

        for (int y = 0; y < height; y++)
        {
            for (int x = 0; x < width; x++)
            {
                Rgba32 pixel = image[x, y];
                byte gray = (byte)((pixel.R + pixel.G + pixel.B) / 3);
                image[x, y] = new Rgba32(gray, gray, gray, pixel.A);
            }
        }
        Console.WriteLine("PixelIterationBenchmark completed");
    }
}
```

The pixel iteration benchmark was implemented in a similar manner with
the same memory diagnostics attributes. The code snippet above
demonstrates the pixel iteration benchmark for ImageSharp, where each
pixel in the image is converted to grayscale. The memory diagnostics
provided by BenchmarkDotNet enabled tracking of the memory consumption
and garbage collection events during the pixel iteration operation,
providing valuable insights into the resource utilization of each
library.

This code exemplifies the approach to memory diagnostics. By annotating
the benchmark class with `[MemoryDiagnoser]`, BenchmarkDotNet
automatically collects data on memory usage—including the number of
garbage collection (GC) events and the total allocated memory during
each benchmarked operation. Similar implementations were done for other
libraries as well.

This level of granularity provided insights that went beyond raw timing
metrics, revealing, for example, that while Emgu CV might be faster in
certain operations, its higher memory consumption could be a concern for
applications running on memory-constrained systems.
