## Pixel Iteration Benchmark Implementation

The pixel iteration benchmark measures the time taken to perform a basic
image processing operation—converting an image to grayscale by iterating
over each pixel. While modern image processing often employs vectorized
operations on entire matrices for efficiency, pixel-by-pixel iteration
remains relevant in several scenarios: when implementing custom filters
with complex logic, when working with specialized pixel formats, or when
memory constraints limit bulk operations. Additionally, this benchmark
provides insight into the underlying performance characteristics of
image libraries even if vectorized alternatives would be preferred in
production environments. By examining the performance of this
fundamental operation, a better understanding of the efficiency
trade-offs in various image processing contexts is achieved.

For ImageSharp, the implementation involves loading the image as an
array of pixels, processing each pixel to compute its grayscale value,
and then updating the image accordingly. The following snippet provides
a glimpse into this process:

``` [Sharp]C
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.Formats.Png;
using SixLabors.ImageSharp.PixelFormats;
public class PixelIterationBenchmark
{
    public static (double warmupTime, double averageTime, double totalTime) RunBenchmark(string imagePath, int iterations)
    {
        long totalElapsedMilliseconds = 0;
        long warmupTime = 0;
        int warmupIterations = 5;
        Stopwatch stopwatch = new Stopwatch();

        // Warm-up phase for pixel iteration
        for (int i = 0; i < warmupIterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image<Rgba32> image = Image.Load<Rgba32>(imagePath))
            {
                int width = image.Width;
                int height = image.Height;
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        Rgba32 pixel = image[x, y];
                        byte gray = (byte)((pixel.R + pixel.G + pixel.B) / 3);
                        image[x, y] = new Rgba32(gray, gray, gray, pixel.A);
                    }
                }
            }
            stopwatch.Stop();
            warmupTime += stopwatch.ElapsedMilliseconds;
        }

        // Main iterations to measure pixel iteration performance
        for (int i = 0; i < iterations; i++)
        {
            stopwatch.Reset();
            stopwatch.Start();
            using (Image<Rgba32> image = Image.Load<Rgba32>(imagePath))
            {
                int width = image.Width;
                int height = image.Height;
                for (int y = 0; y < height; y++)
                {
                    for (int x = 0; x < width; x++)
                    {
                        Rgba32 pixel = image[x, y];
                        byte gray = (byte)((pixel.R + pixel.G + pixel.B) / 3);
                        image[x, y] = new Rgba32(gray, gray, gray, pixel.A);
                    }
                }
            }
            stopwatch.Stop();
            totalElapsedMilliseconds += stopwatch.ElapsedMilliseconds;
            Console.WriteLine($"Iteration {i + 1}: Pixel iteration took {stopwatch.ElapsedMilliseconds} ms");
        }

        double averageTime = totalElapsedMilliseconds / (double)iterations;
        double totalTime = warmupTime + totalElapsedMilliseconds;
        Console.WriteLine($"Warm-up: {warmupTime} ms, Average: {averageTime} ms, Total: {totalTime} ms");

        return (warmupTime, averageTime, totalTime);
    }
}
```

The code measures the performance of a grayscale conversion operation by
iterating over each pixel of an image. As in the image conversion, it
uses a timer (Stopwatch) and divides the process into two phases: a
warm-up phase and a measurement phase. During the warm-up phase, the
image is loaded and processed five times. This phase helps stabilize
performance by mitigating any startup overheads. Each iteration involves
loading the image, iterating over its width and height, reading each
pixel, computing the grayscale value by averaging the red, green, and
blue channels, and assigning the new grayscale value back while
preserving the alpha channel. The use of the `using` statement ensures
that the image is properly disposed after processing.

In the measurement phase, the same processing occurs over a
user-specified number of iterations (100 iterations). After running all
iterations, the code calculates the average time per iteration and the
total time including warm-up. This approach isolates the steady-state
performance from any one-time overhead, resulting in more accurate
measurements that reflect the true cost of pixel-by-pixel manipulations.

The design emphasizes clear resource management, detailed timing, and
separation of initialization costs from the main measurement, which are
crucial when every microsecond of processing time matters in image
manipulation scenarios.

The main focus of the implementation was to capture the interplay
between algorithmic efficiency and system-level resource management.
Every pixel operation is executed in a closed loop, and even minor
inefficiencies can accumulate over hundreds of iterations. The loop
structure is designed and a stopwatch is used to measure elapsed time to
matter of attention that should be paid to details during development.
Because even in high-level libraries such as ImageSharp, every
microsecond counts when processing large images.
