# System Architecture and Design Rationale

The design of the benchmarking framework was guided by the need for
consistency, repeatability, and scientific severity. The system was
architected to support multiple libraries through a common interface,
ensuring that each library’s performance could be measured under
identical conditions. At the core of the design was a two‐phase
benchmarking process: an initial warm-up phase to account for any
initialization overhead, followed by a main test phase where the actual
performance metrics were recorded.

In constructing the system, several important decisions were made.
First, a modular approach was employed, separating the benchmarking
routines into distinct components. This allowed the logic for image
conversion and pixel iteration to be encapsulated into separate classes,
each responsible for executing a series of timed iterations and logging
the results.

``` [Sharp]C
public class ImageConversionBenchmark{
        
        // Benchmarking logic for image conversion
    }
    public class PixelIterationBenchmark{

        // Benchmarking logic for pixel iteration
    }
```

The architecture also included a dedicated component for result
aggregation, which exported data into an Excel file using EPPlus,
thereby facilitating further analysis and visualization.

``` [Sharp]C
using OfficeOpenXml;

    public class ExcelExporter{
        
        // Logic for exporting benchmark results to an Excel sheet in a structured format
    }
```

An essential aspect of the design was the uniformity of testing. Despite
the differences in methods of implementation among the libraries, the
benchmarking framework was designed to abstract away these differences.
Each library was integrated by implementing the same sequence of
operations: reading an image from disk, processing the image (either
converting its format or iterating over its pixels to apply a grayscale
filter), and finally saving the processed image back to disk. This
uniform methodology ensured that the performance comparisons were both
fair and reproducible.

The architecture also accounted for system-level factors such as memory
management and garbage collection. For instance, in languages like C#,
where unmanaged resources must be explicitly disposed of, the design
included rigorous cleanup routines to ensure that each iteration began
with a clean slate. This attention to detail was crucial in obtaining
accurate measurements, as any residual state from previous iterations
could skew the results.

``` [Sharp]C
using BenchmarkDotNet.Attributes;
    using BenchmarkDotNet.Running;

    class Program
    {
        static void Main(string[] args)
        {
            BenchmarkRunner.Run<Benchmarks>();
        }
    }

    [MemoryDiagnoser]
    public class Benchmarks{
        
        [Benchmark]
        public void ImageConversionBenchmark(){
            // Image conversion logic
        }

        [Benchmark]
        public void PixelIterationBenchmark(){
            // Pixel iteration logic
        }
    }
```
